from copy import deepcopy

import numpy as np
import mido
from rich import print as rprint
import midii


def test_sample():
    print(midii.sample.real)
    print(midii.sample.dataset)
    print(midii.sample.simple)


def test_midii_simple_print_tracks():
    ma = midii.MidiFile(midii.sample.simple[0])
    ma.quantize(unit="256")
    ma.print_tracks()


def test_midii_real_print_tracks():
    ma = midii.MidiFile(midii.sample.real[1])
    ma.quantize(unit="256")
    ma.print_tracks(print_note_info=True, track_list=["piano-r"])


def test_mido_dataset_print_tracks():
    ma = mido.MidiFile(midii.sample.dataset[1])
    ma.print_tracks()


def test_midii_print_tracks():
    ma = midii.MidiFile(
        midii.sample.dataset[1], convert_1_to_0=True, lyric_encoding="cp949"
    )
    ma.quantize(unit="32")
    ma.print_tracks()


def test_midii_quantize():
    ma = midii.MidiFile(
        midii.sample.dataset[0], convert_1_to_0=True, lyric_encoding="cp949"
    )
    print(np.mean(np.array(ma.times) % 15))
    ma.quantize(unit="32")
    print(np.mean(np.array(ma.times) % 15))


def test_version():
    from importlib.metadata import version
    import platform

    print("Python Version (concise):", platform.python_version())
    print("mido version:", version("mido"))
    print("rich version:", version("rich"))


def test_midii_print_times():
    ma = midii.MidiFile(
        midii.sample.dataset[0], convert_1_to_0=True, lyric_encoding="cp949"
    )
    ma.print_tracks()


def test_to_json():
    ma = midii.MidiFile(
        midii.sample.dataset[0], convert_1_to_0=True, lyric_encoding="cp949"
    )
    rprint(ma.to_json())
    ma.quantize()
    rprint(ma.to_json())


def test_lyrics():
    ma = midii.MidiFile(
        midii.sample.dataset[0], convert_1_to_0=True, lyric_encoding="cp949"
    )
    print(ma.lyrics)


def test_times():
    ma = midii.MidiFile(
        midii.sample.dataset[0], convert_1_to_0=True, lyric_encoding="cp949"
    )
    print(ma.times)
    ma.quantize()
    print(ma.times)


def test_EF_w_wo():
    ma = midii.MidiFile(
        midii.sample.dataset[0], convert_1_to_0=True, lyric_encoding="cp949"
    )
    ma2 = deepcopy(ma)
    print(np.cumsum(np.array(ma.times, dtype=np.int64))[-10:])
    ma.quantize()
    print(np.cumsum(np.array(ma.times, dtype=np.int64))[-10:])
    ma2.quantize(error_forwarding=False)
    print(np.cumsum(np.array(ma2.times, dtype=np.int64))[-10:])


def test_midi_type():
    ma = midii.MidiFile(
        midii.sample.dataset[0], convert_1_to_0=True, lyric_encoding="cp949"
    )
    print(ma.type)


if __name__ == "__main__":
    # test_sample()
    # test_midii_simple_print_tracks()
    # test_midii_real_print_tracks()
    # test_mido_dataset_print_tracks()
    # test_midii_print_tracks()
    test_midii_quantize()
    # test_midii_print_times()
    # test_version()
    # test_to_json()
    # test_lyrics()
    # test_times()
    # test_EF_w_wo()
    # test_midi_type()



import numpy as np
from numba import njit
from typing import Sequence, Tuple

from .config import DEFAULT_TICKS_PER_BEAT
from .utilities import beat2tick
from .note import Note

_STR2BEAT = {n.value.name_short.split("/")[-1]: n.value.beat for n in Note}


@njit(cache=True, fastmath=True)
def _quantize_w_error_forward(
    ticks: np.ndarray, unit: int
) -> Tuple[np.ndarray, int]:
    """
    Sequential quantization with error forwarding.
    Returns quantized ticks array + final accumulated error.
    """
    quantized_ticks = np.empty_like(ticks)
    err = 0
    for i in range(ticks.size):
        if ticks[i] + err >= 0:
            ticks[i] += err
            err = 0
        r = ticks[i] % unit
        if r * 2 < unit:  # round down
            err += r
            quantized_ticks[i] = ticks[i] - r
        else:  # round up
            err += r - unit
            quantized_ticks[i] = ticks[i] + (unit - r)
    return quantized_ticks, err


def _quantize_wo_error_forward(
    ticks: np.ndarray, unit: int
) -> Tuple[np.ndarray, int]:
    """
    Vectorised midpoint–round-half-up (no error carry).
    """
    q = ticks // unit
    r = ticks - q * unit
    up = r * 2 >= unit
    quantized = (q + up.astype(np.int64)) * unit
    errors = np.where(up, r - unit, r)
    return quantized, int(errors.sum())


def quantize(
    ticks: Sequence[int] | np.ndarray,
    unit: str = "32",
    ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT,
    error_forwarding: bool = True,
) -> Tuple[list[int], int]:
    """
    Parameters
    ----------
    ticks            : iterable of int  –  input delta-ticks
    unit             : e.g. "32"  or numeric beat  or explicit int-ticks
    ticks_per_beat   : resolution (default 480)
    error_forwarding : if True propagate rounding error to next note

    Returns
    -------
    quantized_ticks  : list[int]
    final_error      : int   (may be negative)
    """

    try:
        unit_beat = _STR2BEAT[unit]
    except KeyError:
        raise ValueError(f"unknown unit string {unit!r}")
    unit_tick = beat2tick(unit_beat, ticks_per_beat)
    ticks_arr = np.asarray(ticks, dtype=np.int64)

    if error_forwarding:
        q, err = _quantize_w_error_forward(ticks_arr, unit_tick)
    else:
        q, err = _quantize_wo_error_forward(ticks_arr, unit_tick)

    return q.tolist(), err



from pathlib import Path

sample_path = Path(__file__).parent
dataset = sorted(
    [f for f in (sample_path / "dataset").rglob("*.mid") if f.is_file()]
)
real = sorted(
    [f for f in (sample_path / "real").rglob("*.mid") if f.is_file()]
)
simple = sorted(
    [f for f in (sample_path / "simple").rglob("*.mid") if f.is_file()]
)



import mido

from .config import DEFAULT_TIME_SIGNATURE


def tick2beat(tick, ticks_per_beat):
    return tick / ticks_per_beat


def beat2tick(beat, ticks_per_beat):
    return int(beat * ticks_per_beat)


# Adapted from pretty_midi (utilities.py)
# Source: https://github.com/craffel/pretty-midi/blob/main/pretty_midi/utilities.py
# Original License: MIT
def note_number_to_name(note_number):
    """Convert a MIDI note number to its name, in the format
    ``'(note)(accidental)(octave number)'`` (e.g. ``'C#4'``).

    Parameters
    ----------
    note_number : int
        MIDI note number.  If not an int, it will be rounded.

    Returns
    -------
    note_name : str
        Name of the supplied MIDI note number.

    Notes
    -----
        Thanks to Brian McFee.

    """

    # Note names within one octave
    semis = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]

    # Ensure the note is an int
    note_number = int(round(note_number))

    # Get the semitone and the octave, and concatenate to create the name
    return semis[note_number % 12] + str(note_number // 12 - 1)


def tempo2bpm(tempo, time_signature=DEFAULT_TIME_SIGNATURE):
    return round(mido.tempo2bpm(tempo, time_signature=time_signature))



from collections import namedtuple
from enum import Enum


NoteNamedTuple = namedtuple(
    "NoteNamedTuple", ["beat", "name_eng", "name_kor", "symbol", "name_short"]
)


class Note(Enum):
    WHOLE_NOTE = NoteNamedTuple(4, "whole note", "온음표", "𝅝", "n/1")
    HALF_NOTE = NoteNamedTuple(2, "half note", "2분음표", "𝅗𝅥", "n/2")
    QUARTER_NOTE = NoteNamedTuple(1, "quarter note", "4분음표", "♩", "n/4")
    EIGHTH_NOTE = NoteNamedTuple(0.5, "eighth note", "8분음표", "♪", "n/8")
    SIXTEENTH_NOTE = NoteNamedTuple(
        0.25, "sixteenth note", "16분음표", "𝅘𝅥𝅯", "n/16"
    )
    THIRTY_SECOND_NOTE = NoteNamedTuple(
        0.125, "thirty-second note", "32분음표", "𝅘𝅥𝅰", "n/32"
    )
    SIXTY_FOURTH_NOTE = NoteNamedTuple(
        0.0625, "sixty-fourth note", "64분음표", "𝅘𝅥𝅱", "n/64"
    )
    HUNDRED_TWENTY_EIGHTH_NOTE = NoteNamedTuple(
        0.03125, "hundred twenty-eighth note", "128분음표", "𝅘𝅥𝅲", "n/128"
    )
    TWO_HUNDRED_FIFTY_SIXTH_NOTE = NoteNamedTuple(
        0.015625, "two hundred fifty-sixth note", "256분음표", "𝅘𝅥𝅲", "n/256"
    )


class Rest(Enum):
    WHOLE_REST = NoteNamedTuple(4, "whole rest", "온쉼표", "𝄻", "r/1")
    HALF_REST = NoteNamedTuple(2, "half rest", "2분쉼표", "𝄼", "r/2")
    QUARTER_REST = NoteNamedTuple(1, "quarter rest", "4분쉼표", "𝄽", "r/4")
    EIGHTH_REST = NoteNamedTuple(0.5, "eighth rest", "8분쉼표", "𝄾", "r/8")
    SIXTEENTH_REST = NoteNamedTuple(
        0.25, "sixteenth rest", "16분쉼표", "𝄿", "r/16"
    )
    THIRTY_SECOND_REST = NoteNamedTuple(
        0.125, "thirty-second rest", "32분쉼표", "𝅀", "r/32"
    )
    SIXTY_FOURTH_REST = NoteNamedTuple(
        0.0625, "sixty-fourth rest", "64분쉼표", "𝅁", "r/64"
    )
    HUNDRED_TWENTY_EIGHTH_REST = NoteNamedTuple(
        0.03125, "hundred twenty-eighth rest", "128분쉼표", "𝅂", "r/128"
    )
    TWO_HUNDRED_FIFTY_SIXTH_REST = NoteNamedTuple(
        0.015625, "two hundred fifty-sixth rest", "256분쉼표", "𝅂", "r/256"
    )


class Rest_all(Enum):
    DOTTED_OCTUPLE_WHOLE_REST = NoteNamedTuple(
        48, "dotted octuple whole rest", "점8온쉼표", "𝆶.", "r*8."
    )
    OCTUPLE_WHOLE_REST = NoteNamedTuple(
        32, "octuple whole rest", "8온쉼표", "𝆶", "r*8"
    )
    DOTTED_QUADRUPLE_WHOLE_REST = NoteNamedTuple(
        24, "dotted quadruple whole rest", "점4온쉼표", "𝅜𝅥.", "r*4."
    )
    QUADRUPLE_WHOLE_REST = NoteNamedTuple(
        16, "quadruple whole rest", "4온쉼표", "𝅜", "r*4"
    )
    DOTTED_DOUBLE_REST = NoteNamedTuple(
        12, "dotted double whole rest", "점겹온쉼표", "𝄺.", "r*2."
    )
    DOUBLE_WHOLE_REST = NoteNamedTuple(
        8, "double whole rest", "겹온쉼표", "𝄺", "r*2"
    )
    DOTTED_WHOLE_REST = NoteNamedTuple(
        6, "dotted whole rest", "점온쉼표", "𝄻.", "r/1."
    )
    WHOLE_REST = NoteNamedTuple(4, "whole rest", "온쉼표", "𝄻", "r/1")
    DOTTED_HALF_REST = NoteNamedTuple(
        3, "dotted half rest", "점2분쉼표", "𝄼.", "r/2."
    )
    HALF_REST = NoteNamedTuple(2, "half rest", "2분쉼표", "𝄼", "r/2")
    DOTTED_QUARTER_REST = NoteNamedTuple(
        1.5, "dotted quarter rest", "점4분쉼표", "𝄽.", "r/4."
    )
    QUARTER_REST = NoteNamedTuple(1, "quarter rest", "4분쉼표", "𝄽", "r/4")
    DOTTED_EIGHTH_REST = NoteNamedTuple(
        0.75, "dotted eighth rest", "점8분쉼표", "𝄾.", "r/8."
    )
    EIGHTH_REST = NoteNamedTuple(0.5, "eighth rest", "8분쉼표", "𝄾", "r/8")
    DOTTED_SIXTEENTH_REST = NoteNamedTuple(
        0.375, "dotted sixteenth rest", "점16분쉼표", "𝄿.", "r/16."
    )
    SIXTEENTH_REST = NoteNamedTuple(
        0.25, "sixteenth rest", "16분쉼표", "𝄿", "r/16"
    )
    DOTTED_THIRTY_SECOND_REST = NoteNamedTuple(
        0.1875, "dotted thirty-second rest", "점32분쉼표", "𝅀.", "r/32."
    )
    THIRTY_SECOND_REST = NoteNamedTuple(
        0.125, "thirty-second rest", "32분쉼표", "𝅀", "r/32"
    )
    DOTTED_SIXTY_FOURTH_REST = NoteNamedTuple(
        0.09375, "dotted sixty-fourth rest", "점64분쉼표", "𝅁.", "r/64."
    )
    SIXTY_FOURTH_REST = NoteNamedTuple(
        0.0625, "sixty-fourth rest", "64분쉼표", "𝅁", "r/64"
    )
    DOTTED_HUNDRED_TWENTY_EIGHTH_REST = NoteNamedTuple(
        0.046875,
        "dotted hundred twenty-eighth rest",
        "점128분쉼표",
        "𝅂.",
        "r/128.",
    )
    HUNDRED_TWENTY_EIGHTH_REST = NoteNamedTuple(
        0.03125, "hundred twenty-eighth rest", "128분쉼표", "𝅂", "r/128"
    )
    DOTTED_TWO_HUNDRED_FIFTY_SIXTH_REST = NoteNamedTuple(
        0.0234375,
        "dotted two hundred fifty-sixth rest",
        "점256분쉼표",
        "𝅂.",
        "r/256.",
    )
    TWO_HUNDRED_FIFTY_SIXTH_REST = NoteNamedTuple(
        0.015625, "two hundred fifty-sixth rest", "256분쉼표", "𝅂", "r/256"
    )


class Note_all(Enum):
    DOTTED_OCTUPLE_WHOLE_NOTE = NoteNamedTuple(
        48, "dotted octuple whole note", "점8온음표", "𝆶.", "n*8."
    )
    OCTUPLE_WHOLE_NOTE = NoteNamedTuple(
        32, "octuple whole note", "8온음표", "𝆶", "n*8"
    )
    DOTTED_QUADRUPLE_WHOLE_NOTE = NoteNamedTuple(
        24, "dotted quadruple whole note", "점4온음표", "𝅜𝅥.", "n*4."
    )
    QUADRUPLE_WHOLE_NOTE = NoteNamedTuple(
        16, "quadruple whole note", "4온음표", "𝅜", "n*4"
    )
    DOTTED_DOUBLE_WHOLE_NOTE = NoteNamedTuple(
        12, "dotted double whole note", "점겹온음표", "𝅜.", "n*2."
    )
    DOUBLE_WHOLE_NOTE = NoteNamedTuple(
        8, "double whole note", "겹온음표", "𝅜", "n*2"
    )
    DOTTED_WHOLE_NOTE = NoteNamedTuple(
        6, "dotted whole note", "점온음표", "𝅝.", "n/1."
    )
    WHOLE_NOTE = NoteNamedTuple(4, "whole note", "온음표", "𝅝", "n/1")
    DOTTED_HALF_NOTE = NoteNamedTuple(
        3, "dotted half note", "점2분음표", "♩.", "n/2."
    )
    HALF_NOTE = NoteNamedTuple(2, "half note", "2분음표", "𝅗𝅥", "n/2")
    DOTTED_QUARTER_NOTE = NoteNamedTuple(
        1.5, "dotted quarter note", "점4분음표", "♩.", "n/4."
    )
    QUARTER_NOTE = NoteNamedTuple(1, "quarter note", "4분음표", "♩", "n/4")
    DOTTED_EIGHTH_NOTE = NoteNamedTuple(
        0.75, "dotted eighth note", "점8분음표", "♪.", "n/8."
    )
    EIGHTH_NOTE = NoteNamedTuple(0.5, "eighth note", "8분음표", "♪", "n/8")
    DOTTED_SIXTEENTH_NOTE = NoteNamedTuple(
        0.375, "dotted sixteenth note", "점16분음표", "𝅘𝅥𝅯.", "n/16."
    )
    SIXTEENTH_NOTE = NoteNamedTuple(
        0.25, "sixteenth note", "16분음표", "𝅘𝅥𝅯", "n/16"
    )
    DOTTED_THIRTY_SECOND_NOTE = NoteNamedTuple(
        0.1875, "dotted thirty-second note", "점32분음표", "𝅘𝅥𝅰.", "n/32."
    )
    THIRTY_SECOND_NOTE = NoteNamedTuple(
        0.125, "thirty-second note", "32분음표", "𝅘𝅥𝅰", "n/32"
    )
    DOTTED_SIXTY_FOURTH_NOTE = NoteNamedTuple(
        0.09375, "dotted sixty-fourth note", "점64분음표", "𝅘𝅥𝅱.", "n/64."
    )
    SIXTY_FOURTH_NOTE = NoteNamedTuple(
        0.0625, "sixty-fourth note", "64분음표", "𝅘𝅥𝅱", "n/64"
    )
    DOTTED_HUNDRED_TWENTY_EIGHTH_NOTE = NoteNamedTuple(
        0.046875,
        "dotted hundred twenty-eighth note",
        "점128분음표",
        "𝅘𝅥𝅲.",
        "n/128.",
    )
    HUNDRED_TWENTY_EIGHTH_NOTE = NoteNamedTuple(
        0.03125, "hundred twenty-eighth note", "128분음표", "𝅘𝅥𝅲", "n/128"
    )
    DOTTED_TWO_HUNDRED_FIFTY_SIXTH_NOTE = NoteNamedTuple(
        0.0234375,
        "dotted two hundred fifty-sixth note",
        "점256분음표",
        "𝅘𝅥𝅲.",
        "n/256.",
    )
    TWO_HUNDRED_FIFTY_SIXTH_NOTE = NoteNamedTuple(
        0.015625, "two hundred fifty-sixth note", "256분음표", "𝅘𝅥𝅲", "n/256"
    )



import mido
from rich import print as rprint
from rich.console import Console
from rich.panel import Panel

from .note import Note
from .messages import (
    MessageAnalyzer_text,
    MessageAnalyzer_set_tempo,
    MessageAnalyzer_end_of_track,
    MessageAnalyzer_key_signature,
    MessageAnalyzer_time_signature,
    MessageAnalyzer,
    MessageAnalyzer_note_on,
    MessageAnalyzer_note_off,
    MessageAnalyzer_lyrics,
)
from .config import (
    DEFAULT_TICKS_PER_BEAT,
    DEFAULT_TEMPO,
    DEFAULT_TIME_SIGNATURE,
)
from .quantize import quantize


class MidiFile(mido.MidiFile):
    def __init__(
        self,
        filename=None,
        file=None,
        type=1,
        ticks_per_beat=DEFAULT_TICKS_PER_BEAT,
        charset="latin1",
        debug=False,
        clip=False,
        tracks=None,
        convert_1_to_0=False,
        lyric_encoding="latin1",
    ):
        super().__init__(
            filename=filename,
            file=file,
            type=type,
            ticks_per_beat=ticks_per_beat,
            charset=charset,
            debug=debug,
            clip=clip,
            tracks=tracks,
        )

        self.lyric_encoding = lyric_encoding
        self.convert_1_to_0 = convert_1_to_0

        if self.type == 1 and self.convert_1_to_0:
            self.tracks = [self.merged_track]
            self.type = 0

    def quantize(
        self,
        unit="32",
        error_forwarding=True,
        targets=["note_on", "note_off", "lyrics"],
    ):
        if self.type == 0:
            quantized_ticks, error = quantize(
                ticks=self.times,
                unit=unit,
                ticks_per_beat=self.ticks_per_beat,
                error_forwarding=error_forwarding,
            )
            quantized_ticks_iter = iter(quantized_ticks)
            for msg in self.tracks[0]:
                if msg.type in targets:
                    msg.time = next(quantized_ticks_iter)
        elif self.type == 1:
            for track, track_times in zip(self.tracks, self.times):
                quantized_ticks, error = quantize(
                    ticks=track_times,
                    unit=unit,
                    ticks_per_beat=self.ticks_per_beat,
                    error_forwarding=error_forwarding,
                )
                quantized_ticks_iter = iter(quantized_ticks)
                for msg in track:
                    if msg.type in targets:
                        msg.time = next(quantized_ticks_iter)
        else:  # type == 2
            raise NotImplementedError

    def _print_note_num(self, note_num, tempo, time_signature):
        color = "color(240)" if note_num == 0 else "color(47)"
        bpm = round(mido.tempo2bpm(tempo, time_signature=time_signature))
        info = f"[bold {color}]Total item num of BPM({bpm}): " + f"{note_num}"
        Console().rule(info, style=f"{color}")

    def _print_tracks(
        self,
        track,
        track_limit=None,
        print_note=True,
        print_time=True,
        print_note_info=False,
    ):
        tempo = DEFAULT_TEMPO
        time_signature = DEFAULT_TIME_SIGNATURE
        current_time = 0
        note_address = 0
        note_num = 0
        first_tempo = True
        prev_tempo = None
        note_queue = {}
        if track_limit is None:
            track_limit = float("inf")
        total_time = 0
        for i, msg in enumerate(track):
            if i > track_limit:
                break
            total_time += msg.time
            current_time += mido.tick2second(
                msg.time,
                ticks_per_beat=self.ticks_per_beat,
                tempo=tempo,
            )
            kwarg = {
                "msg": msg,
                "ticks_per_beat": self.ticks_per_beat,
                "tempo": tempo,
                "index": i,
                "current_time": current_time,
                "print_time": print_time,
            }
            if msg.type == "note_on":
                ma = MessageAnalyzer_note_on(
                    **kwarg,
                    print_note=print_note,
                    print_note_info=print_note_info,
                    note_queue=note_queue,
                )
                note_address = ma.addr
            elif msg.type == "note_off":
                ma = MessageAnalyzer_note_off(
                    **kwarg,
                    print_note=print_note,
                    print_note_info=print_note_info,
                    note_queue=note_queue,
                )
            elif msg.type == "lyrics":
                ma = MessageAnalyzer_lyrics(
                    **kwarg,
                    print_note=print_note,
                    print_note_info=print_note_info,
                    encoding=self.lyric_encoding,
                    note_address=note_address,
                )
            elif msg.type == "text" or msg.type == "track_name":
                ma = MessageAnalyzer_text(
                    **kwarg,
                    encoding=self.lyric_encoding,
                )
            elif msg.type == "set_tempo":
                if not first_tempo and self.convert_1_to_0:
                    self._print_note_num(note_num, tempo, time_signature)
                first_tempo = False
                tempo = msg.tempo
                ma = MessageAnalyzer_set_tempo(
                    **kwarg,
                    time_signature=time_signature,
                )
                if prev_tempo is None:
                    prev_tempo = tempo
                if note_num:
                    prev_tempo = tempo
                    note_num = 0
                else:
                    tempo = prev_tempo
            elif msg.type == "end_of_track":
                if self.convert_1_to_0:
                    self._print_note_num(note_num, tempo, time_signature)
                ma = MessageAnalyzer_end_of_track(**kwarg)
            elif msg.type == "key_signature":
                ma = MessageAnalyzer_key_signature(**kwarg)
            elif msg.type == "time_signature":
                time_signature = (msg.numerator, msg.denominator)
                ma = MessageAnalyzer_time_signature(**kwarg)
            else:
                ma = MessageAnalyzer(**kwarg)

            _str = str(ma)
            if _str:
                rprint(_str)

            if msg.type in ["note_on", "note_off", "lyrics"]:
                note_num += 1

        current_time = mido.tick2second(
            total_time,
            ticks_per_beat=self.ticks_per_beat,
            tempo=tempo,
        )
        rprint("Total secs/time: " + f"{current_time}/{total_time}")

    def _panel(self):
        # meta information of midi file
        header_style = "black on white blink"
        header_info = "\n".join(
            [
                f"[{header_style}]mid file type: {self.type}",
                f"ticks per beat: {self.ticks_per_beat}",
                f"total duration: {self.length}[/{header_style}]",
            ]
        )
        return Panel(
            header_info,
            title="[MIDI File Header]",
            subtitle=f"{self.filename}",
            style=f"{header_style}",
            border_style=f"{header_style}",
        )

    def print_tracks(
        self,
        track_limit=None,
        print_note=True,
        print_time=True,
        print_note_info=False,
        track_list=None,
    ):
        if track_limit is None:
            track_limit = float("inf")
        rprint(self._panel())

        _style_track_line = "#ffffff on #4707a8"
        for i, track in enumerate(self.tracks):
            Console().rule(
                f'[{_style_track_line}]Track {i}: "{track.name}"'
                f"[/{_style_track_line}]",
                style=f"{_style_track_line}",
            )
            if track_list is None or track.name in track_list:
                self._print_tracks(
                    track,
                    track_limit=track_limit,
                    print_note=print_note,
                    print_time=print_time,
                    print_note_info=print_note_info,
                )

    def _get_note_data(
        self, msg, note_map, time_current, tempo, lyric, time_format
    ):
        note_data = {
            "start": None,
            "end": None,
            "duration": None,
            "pitch": None,
            "lyric": None,
        }
        note_data = note_data.copy()
        try:
            time_note_on = note_map[msg.note]
        except KeyError:
            note_data["pitch"] = 0
            time_note_on = time_current - msg.time
        else:
            note_data["pitch"] = msg.note
            del note_map[msg.note]
        duration = time_current - time_note_on
        if time_format == "ticks":
            note_data["start"] = time_note_on
            note_data["end"] = time_current
            note_data["duration"] = duration
        elif time_format == "seconds":
            note_data["start"] = mido.tick2second(
                time_note_on,
                ticks_per_beat=self.ticks_per_beat,
                tempo=tempo,
            )
            note_data["end"] = mido.tick2second(
                time_current,
                ticks_per_beat=self.ticks_per_beat,
                tempo=tempo,
            )
            note_data["duration"] = mido.tick2second(
                duration,
                ticks_per_beat=self.ticks_per_beat,
                tempo=tempo,
            )
        else:
            raise ValueError
        note_data["lyric"] = lyric
        return note_data

    def to_json(
        self,
        time_format="ticks",  # "seconds", "ticks"
    ):
        if self.type == 1 and not self.convert_1_to_0:
            raise RuntimeError
        tempo = DEFAULT_TEMPO
        time_current = 0
        lyric = ""
        result = []
        note_map = {}
        for msg in self.tracks[0]:
            time_current += msg.time
            if msg.type == "set_tempo":
                tempo = msg.tempo
            elif msg.type == "lyrics":
                lyric += MessageAnalyzer_lyrics(
                    msg=msg,
                    encoding=self.lyric_encoding,
                ).lyric
            elif msg.type == "note_on":
                note_map[msg.note] = time_current
            elif msg.type == "note_off":
                result.append(
                    self._get_note_data(
                        msg, note_map, time_current, tempo, lyric, time_format
                    )
                )
                lyric = ""

        return result

    @property
    def times(self):
        if self.type == 0:
            return [
                msg.time
                for msg in self.tracks[0]
                if msg.type in ["note_on", "note_off", "lyrics"]
            ]
        elif self.type == 1:
            return [
                [
                    msg.time
                    for msg in track
                    if msg.type in ["note_on", "note_off", "lyrics"]
                ]
                for track in self.tracks
            ]
        else:  # type == 2
            raise NotImplementedError

    @property
    def lyrics(self):
        lyrics = ""
        if self.type == 0:
            for msg in self.tracks[0]:
                if msg.type == "lyrics":
                    lyrics += MessageAnalyzer_lyrics(
                        msg=msg,
                        encoding=self.lyric_encoding,
                    ).lyric
        elif self.type == 1:
            for track in self.tracks:
                for msg in track:
                    if msg.type == "lyrics":
                        lyrics += MessageAnalyzer_lyrics(
                            msg=msg,
                            encoding=self.lyric_encoding,
                        ).lyric
        elif self.type == 2:
            raise NotImplementedError
        return lyrics



import string

import mido

from .note import (
    Note_all,
    Rest_all,
)
from .config import (
    DEFAULT_TICKS_PER_BEAT,
    DEFAULT_TEMPO,
    DEFAULT_TIME_SIGNATURE,
    COLOR,
)
from .utilities import tick2beat, note_number_to_name, tempo2bpm


class MessageAnalyzer:
    def __init__(
        self,
        msg,
        ticks_per_beat=DEFAULT_TICKS_PER_BEAT,
        tempo=DEFAULT_TEMPO,
        index=0,
        current_time=0,
        print_time=True,
    ):
        self.msg = msg
        self.ticks_per_beat = ticks_per_beat
        self.tempo = tempo
        self.index = index
        self.current_time = current_time
        self.print_time = print_time

    def str_type(self):
        return f"[black on white]\\[{self.msg.type}][/black on white]"

    def str_time(self):
        if not self.msg.time:
            return ""
        style_main = "#ffffff"
        style_sub = "white"
        time = mido.tick2second(
            self.msg.time,
            ticks_per_beat=self.ticks_per_beat,
            tempo=self.tempo,
        )
        return " ".join(
            [
                f"[{style_main}]{time:4.2f}[/{style_main}]"
                + f"[{style_sub}]/{self.current_time:6.2f}[/{style_sub}]",
                f"[{style_sub}]time=[/{style_sub}]"
                + f"[{style_main}]{self.msg.time:<3}[/{style_main}]",
            ]
        )

    def str_format(self, head="", body=""):
        _str_time = "" if not self.print_time else self.str_time()
        _str_idx = f"[color(244)]{self.index:4}[/color(244)]"
        return " ".join([s for s in [_str_idx, head, _str_time, body] if s])

    def __str__(self):
        return self.str_format(
            head=self.str_type(),
            body=f"[color(250)]{self.msg}[/color(250)]",
        )


class MessageAnalyzer_set_tempo(MessageAnalyzer):
    def __init__(
        self,
        msg,
        ticks_per_beat=DEFAULT_TICKS_PER_BEAT,
        tempo=DEFAULT_TEMPO,
        index=0,
        current_time=0,
        print_time=True,
        time_signature=DEFAULT_TIME_SIGNATURE,
    ):
        super().__init__(
            msg,
            ticks_per_beat,
            tempo=tempo,
            index=index,
            current_time=current_time,
            print_time=print_time,
        )
        self.time_signature = time_signature

    def __str__(self):
        bpm = tempo2bpm(self.msg.tempo, time_signature=self.time_signature)
        return self.str_format(
            head=self.str_type(),
            body=f"[white]BPM=[/white][color(190)]{bpm}"
            + f"({self.msg.tempo})[/color(190)]",
        )


class MessageAnalyzer_key_signature(MessageAnalyzer):
    def __str__(self):
        return self.str_format(head=self.str_type(), body=self.msg.key)


class MessageAnalyzer_end_of_track(MessageAnalyzer):
    def __str__(self):
        return self.str_format(head=self.str_type())


class MessageAnalyzer_time_signature(MessageAnalyzer):
    def __str__(self):
        result = self.str_format(
            head=self.str_type(),
            body=f"{self.msg.numerator}/{self.msg.denominator}",
        )
        return result


class MessageAnalyzer_text(MessageAnalyzer):
    def __init__(
        self,
        msg,
        ticks_per_beat=DEFAULT_TICKS_PER_BEAT,
        tempo=DEFAULT_TEMPO,
        index=0,
        current_time=0,
        print_time=True,
        encoding="latin-1",
    ):
        super().__init__(
            msg,
            ticks_per_beat,
            tempo=tempo,
            index=index,
            current_time=current_time,
            print_time=print_time,
        )
        self.text = self.decode(encoding=encoding)

    def decode(self, encoding):
        return self.msg.bin()[3:].decode(encoding).strip()

    def __str__(self):
        return self.str_format(head=self.str_type(), body=self.text)


class MessageAnalyzer_SoundUnit(MessageAnalyzer):
    def __init__(
        self,
        msg,
        ticks_per_beat=DEFAULT_TICKS_PER_BEAT,
        tempo=DEFAULT_TEMPO,
        index=0,
        current_time=0,
        print_time=True,
        print_note=True,
        print_note_info=False,
        note_queue=None,
    ):
        super().__init__(
            msg,
            ticks_per_beat,
            tempo=tempo,
            index=index,
            current_time=current_time,
            print_time=print_time,
        )
        if note_queue is None:
            self.note_queue = {}
        else:
            self.note_queue = note_queue
        self.print_note = print_note
        self.print_note_info = print_note_info

    def note_queue_find(self, value):
        for k, v in self.note_queue.items():
            if v == value:
                return k
        return None

    def note_queue_alloc(self):
        address = 0
        while True:
            try:
                self.note_queue[address]
                address += 1
            except KeyError:
                return address

    def closest_note(self, tick, as_rest=False):
        if tick == 0:
            return None, None
        beat = tick2beat(tick, self.ticks_per_beat)
        min_error = float("inf")
        quantized_note = None
        note_enum = Rest_all if as_rest else Note_all
        for note in note_enum:
            error = note.value.beat - beat
            if abs(error) < min_error:
                min_error = error
                quantized_note = note.value
        return min_error, quantized_note

    def str_quantization(
        self, error, real_beat, quantized_note, quantization_color="color(85)"
    ):
        if error is None:
            return ""
        if error == 0:
            err_msg = ""
        else:
            err_msg = (
                f"[red]-{float(real_beat):.3}[/red]"
                + f"[#ff0000]={error}[/#ff0000]"
            )
        return (
            f"[{quantization_color}]"
            + f"{quantized_note.symbol:2}{quantized_note.name_short}"
            + f"[/{quantization_color}] "
            + f"[color(249)]{float(quantized_note.beat):.3}b[/color(249)]"
            + err_msg
        )

    def str_note(self, note):
        return f"{note_number_to_name(note):>3}({note:2})"


class MessageAnalyzer_note_on(MessageAnalyzer_SoundUnit):
    def __init__(
        self,
        msg,
        ticks_per_beat=DEFAULT_TICKS_PER_BEAT,
        tempo=DEFAULT_TEMPO,
        index=0,
        current_time=0,
        print_time=True,
        print_note=True,
        print_note_info=False,
        note_queue=None,
    ):
        super().__init__(
            msg,
            ticks_per_beat,
            tempo=tempo,
            index=index,
            current_time=current_time,
            print_time=print_time,
            print_note=print_note,
            print_note_info=print_note_info,
            note_queue=note_queue,
        )
        self.addr = self.alloc_note(self.msg.note)

    def alloc_note(self, note):
        note_address = self.note_queue_alloc()
        self.note_queue[note_address] = note
        return note_address

    def __str__(self):
        error, quantized_note = self.closest_note(self.msg.time, as_rest=True)
        _str_quantization = ""
        if error is not None and self.print_note_info:
            _str_quantization = self.str_quantization(
                round(error, 3),
                tick2beat(self.msg.time, self.ticks_per_beat),
                quantized_note,
            )
        color = f"color({COLOR[self.addr % len(COLOR)]})"
        note_msg = f"[{color}]┌{self.str_note(self.msg.note)}┐[/{color}]"
        result = ""
        if self.print_note:
            result = self.str_format(head=note_msg, body=_str_quantization)
        return result


class MessageAnalyzer_note_off(MessageAnalyzer_SoundUnit):
    def __init__(
        self,
        msg,
        ticks_per_beat=DEFAULT_TICKS_PER_BEAT,
        tempo=DEFAULT_TEMPO,
        index=0,
        current_time=0,
        print_time=True,
        print_note=True,
        print_note_info=False,
        note_queue=None,
    ):
        super().__init__(
            msg,
            ticks_per_beat,
            tempo=tempo,
            index=index,
            current_time=current_time,
            print_time=print_time,
            print_note=print_note,
            print_note_info=print_note_info,
            note_queue=note_queue,
        )
        self.addr = self.free_note(self.msg.note)

    def free_note(self, note):
        addr = self.note_queue_find(note)
        if addr is not None:
            del self.note_queue[addr]
        return addr

    def __str__(self):
        color = (
            None
            if self.addr is None
            else f"color({COLOR[self.addr % len(COLOR)]})"
        )
        error, quantized_note = self.closest_note(
            self.msg.time, as_rest=True if self.addr is None else False
        )
        if color:
            _note_info = self.str_note(self.msg.note)
            _str_note_off_info = f"[{color}]└{_note_info}┘[/{color}]"
        else:
            symbol = quantized_note.symbol if quantized_note else "0"
            _str_note_off_info = f"[#ffffff]{symbol:^9}[/#ffffff]"
        _str_quantization = ""
        if error is not None and self.print_note_info:
            _str_quantization = self.str_quantization(
                round(error, 3),
                tick2beat(self.msg.time, self.ticks_per_beat),
                quantized_note,
            )
        result = ""
        if self.print_note:
            result = self.str_format(
                head=_str_note_off_info,
                body=_str_quantization,
            )
        return result


class MessageAnalyzer_lyrics(MessageAnalyzer_SoundUnit, MessageAnalyzer_text):
    def __init__(
        self,
        msg,
        ticks_per_beat=DEFAULT_TICKS_PER_BEAT,
        tempo=DEFAULT_TEMPO,
        index=0,
        current_time=0,
        print_time=True,
        print_note=True,
        print_note_info=False,
        encoding="latin-1",
        note_address=0,
    ):
        self.msg = msg
        self.ticks_per_beat = ticks_per_beat
        self.tempo = tempo
        self.index = index
        self.current_time = current_time
        self.print_time = print_time
        self.print_note = print_note
        self.print_note_info = print_note_info
        self.lyric = self.decode(encoding=encoding)
        if not self.lyric:
            self.lyric = " "
        self.note_address = note_address

    def is_alnumpunc(self, s):
        candidate = (
            string.ascii_letters + string.digits + string.punctuation + " "
        )
        for c in s:
            if c not in candidate:
                return False
        return True

    def __str__(self):
        style_lyric = "bold #98ff29"
        style_border = f"color({COLOR[self.note_address % len(COLOR)]})"

        border = f"[{style_border}]│[/{style_border}]"
        _str_lyric = (
            f"{self.lyric:^7}"
            if self.is_alnumpunc(self.lyric)
            else f"{self.lyric:^6}"
        )

        error, quantized_note = self.closest_note(self.msg.time)
        info_quantization = ""
        if error is not None and self.print_note_info:
            info_quantization = self.str_quantization(
                round(error, 3),
                tick2beat(self.msg.time, self.ticks_per_beat),
                quantized_note,
            )
        head = (
            border
            + f"[{style_lyric}]"
            + _str_lyric
            + f"[/{style_lyric}]"
            + border
        )
        result = ""
        if self.print_note:
            result = self.str_format(
                head=head,
                body=info_quantization,
            )
        return result



import midii
import mido
from numba import njit


def test_sample():
    print(midii.sample.real)
    print(midii.sample.dataset)
    print(midii.sample.simple)


def test_midii_simple_print_tracks():
    ma = midii.MidiFile(midii.sample.simple[0])
    ma.quantize(unit="256")
    ma.print_tracks()


def test_midii_real_print_tracks():
    ma = midii.MidiFile(midii.sample.real[1])
    ma.quantize(unit="256")
    ma.print_tracks(print_note_info=False, track_list=["piano-r"])


def test_mido_dataset_print_tracks():
    ma = mido.MidiFile(midii.sample.dataset[1])
    ma.print_tracks()


def test_midii_print_tracks():
    ma = midii.MidiFile(
        midii.sample.dataset[1], convert_1_to_0=True, lyric_encoding="cp949"
    )
    # ma.quantize(unit="32")
    ma.print_tracks(
        track_limit=None,
        track_list=None,
        print_note_info=False,
    )


def test_midii_quantization():
    ma = midii.MidiFile(
        midii.sample.dataset[0],
        lyric_encoding="cp949",
        # midii.sample.dataset[0], convert_1_to_0=True, lyric_encoding="cp949"
    )
    ma.quantize(unit="32", error_forwarding=False)
    ma.print_tracks(
        track_limit=None,
        track_list=None,
        print_note_info=False,
    )


def test_version():
    from importlib.metadata import version
    import platform

    pkgs = [
        "mido",
        "rich",
        "ipykernel",
        "matplotlib",
        "pytest",
        "numpy",
        "numba",
    ]
    for pkg in pkgs:
        print(f"{pkg} version:", version(pkg))
    print("Python version:", platform.python_version())

    # mido version: 1.3.3
    # rich version: 14.0.0
    # ipykernel version: 6.29.5
    # matplotlib version: 3.10.1
    # pytest version: 8.3.5
    # numpy version: 2.2.5
    # numba version: 0.61.2
    # Python version: 3.13.1

    # print("mido version:", version("mido"))
    # print("numpy version:", version("numpy"))
    # print("rich version:", version("rich"))
    # print("numba version:", version("numba"))


def test_midii_print_times():
    ma = midii.MidiFile(
        midii.sample.dataset[0], convert_1_to_0=True, lyric_encoding="cp949"
    )
    # ma.print_tracks()
    print(ma.times)
    ma.quantize(unit="64")
    print(ma.times)


def test_standalone_quantize():
    ma = midii.MidiFile(
        midii.sample.dataset[0], convert_1_to_0=True, lyric_encoding="cp949"
    )
    subset = slice(0, 70)
    subset_last = slice(-33, None)
    times_q32, error_q32 = midii.quantize(
        ma.times, unit="32", ticks_per_beat=ma.ticks_per_beat
    )
    times_q64, error_q64 = midii.quantize(
        ma.times, unit="64", ticks_per_beat=ma.ticks_per_beat
    )
    times_q128, error_q128 = midii.quantize(
        ma.times, unit="128", ticks_per_beat=ma.ticks_per_beat
    )
    # print(ma.times[subset])
    print(ma.times[subset_last])
    ma.quantize()
    # print(ma.times[subset])
    # print(times_q32[subset])
    print(ma.times[subset_last])
    print(times_q32[subset_last])
    # print(times_q64[subset], error_q64)
    # print(times_q128[subset], error_q128)
    # print(times_q64[subset_last])
    # print(times_q128[subset_last])


def test_divmod(t, u):
    if False:
        # $ hyperfine --warmup 10 -r 200 "python test.py"
        # Benchmark 1: python test.py
        #   Time (mean ± σ):     228.3 ms ±   8.0 ms    [User: 103.5 ms, System: 90.9 ms]
        #   Range (min … max):   215.8 ms … 262.7 ms    200 runs
        q, r = divmod(t, u)
    else:
        # hyperfine --warmup 10 -r 200 "python test.py"
        # Benchmark 1: python test.py
        #   Time (mean ± σ):     229.2 ms ±   8.5 ms    [User: 104.3 ms, System: 89.7 ms]
        #   Range (min … max):   215.4 ms … 275.7 ms    200 runs
        q = t // u
        r = t - q * u  # r = t % u
    return q, r


def test_remainder():
    # for i in range(100_000):
    # r = i % 7
    for i in range(100_000):
        q = i // 7
        r = i - q * 7
    return r


@njit(cache=True, fastmath=True)
def test_remainder_numba():
    for i in range(100_000):
        r = i % 7
    # for i in range(100_000):
    #     q = i // 7
    #     r = i - q * 7
    return r


if __name__ == "__main__":
    # test_sample()
    # test_midii_simple_print_tracks()
    # test_midii_real_print_tracks()
    # test_mido_dataset_print_tracks()
    # test_midii_print_tracks()
    # test_midii_quantization()
    # test_version()
    # test_midii_print_times()
    # test_standalone_quantize()
    # test_divmod(100, 18)
    # test_remainder()
    test_remainder_numba()
